import json
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Type, Union

import dacite

from nzshm_model.psha_adapter import PshaAdapterInterface


@dataclass
class Branch:
    gsim_name: str
    gsim_args: Dict[str, Any] = field(default_factory=dict)
    weight: float = 1.0


@dataclass
class BranchSet:
    tectonic_region_type: str
    branches: List[Branch] = field(default_factory=list)


@dataclass
class GMCMLogicTree:
    version: str = ''
    title: str = ''
    branch_sets: List[BranchSet] = field(default_factory=list)

    def __post_init__(self):
        self._fix_args()

    def _fix_args(self) -> 'GMCMLogicTree':
        """Replace string representations of numeric arguments with floats"""

        def is_number(value):
            return value.replace("-", "").replace(".", "").replace("e", "").isnumeric()

        for branch_set in self.branch_sets:
            for branch in branch_set.branches:
                for k, v in branch.gsim_args.items():
                    if (isinstance(v, str)) and (is_number(v)):
                        branch.gsim_args[k] = float(v)

        return self

    @staticmethod
    def from_dict(data: Dict) -> 'GMCMLogicTree':
        return dacite.from_dict(data_class=GMCMLogicTree, data=data, config=dacite.Config(strict=True))

    @staticmethod
    def from_json(json_path: Union[Path, str]) -> 'GMCMLogicTree':
        with Path(json_path).open() as jsonfile:
            data = json.load(jsonfile)
        return GMCMLogicTree.from_dict(data)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def psha_adapter(self, provider: Type[PshaAdapterInterface], **kwargs):
        return provider(gmcm_logic_tree=self)



"""
test migration from slt v1 to v2
"""
import pytest

import nzshm_model
from nzshm_model.source_logic_tree import SourceLogicTree
from nzshm_model.source_logic_tree.version2 import Branch


## three example filter functions
def unscaled_filter_str(obj):
    return "s1.0" in [str(v) for v in obj.values]


def unscaled_filter(obj):
    for v in obj.values:
        if v.name == 's' and v.value == 1.0:  # moment rate scaling
            return True
    return False


def geodetic_filter(obj):
    for v in obj.values:
        if v.long_name == 'deformation model' and v.value == "geodetic":
            return True


@pytest.fixture(scope='module')
def full_slt():
    yield nzshm_model.get_model_version('NSHM_v1.0.4').source_logic_tree()


def test_filter_fn_on_logic_tree_branches(full_slt):
    # functional style
    assert len([filt_branch for filt_branch in filter(lambda obj: obj.fslt.short_name == 'HIK', full_slt)]) == 9
    assert len([filt_branch for filt_branch in filter(unscaled_filter_str, full_slt)]) == 16
    assert len([filt_branch for filt_branch in filter(unscaled_filter, full_slt)]) == 16
    assert len([filt_branch for filt_branch in filter(geodetic_filter, full_slt)]) == 18
    assert len([filt_branch for filt_branch in filter(unscaled_filter, filter(geodetic_filter, full_slt))]) == 6


def test_list_comprehension_filtering_logic_tree_branches(full_slt):
    # old-skool pythonic style
    assert len([filt_branch for filt_branch in full_slt if filt_branch.fslt.short_name == "HIK"]) == 9
    assert len([filt_branch for filt_branch in full_slt if unscaled_filter_str(filt_branch)]) == 16
    assert len([filt_branch for filt_branch in full_slt if unscaled_filter(filt_branch)]) == 16
    assert len([filt_branch for filt_branch in full_slt if geodetic_filter(filt_branch)]) == 18
    assert (
        len([filt_branch for filt_branch in full_slt if unscaled_filter(filt_branch) and geodetic_filter(filt_branch)])
        == 6
    )


def test_generator_filtering_logic_tree_branches(full_slt):
    # most memory efficient & pythonic
    assert sum((1 for filt_branch in full_slt if filt_branch.fslt.short_name == "HIK")) == 9
    assert sum(1 for filt_branch in full_slt if unscaled_filter_str(filt_branch)) == 16
    assert sum(1 for filt_branch in full_slt if unscaled_filter(filt_branch)) == 16
    assert sum(1 for filt_branch in full_slt if geodetic_filter(filt_branch)) == 18
    assert sum(1 for filt_branch in full_slt if unscaled_filter(filt_branch) and geodetic_filter(filt_branch)) == 6


def test_build_slt_from_filtered_slt(full_slt):

    flt = (filt_branch for filt_branch in full_slt if filt_branch.fslt.short_name == "HIK")
    slt = SourceLogicTree.from_branches(flt)
    print(slt)
    assert len(slt.fault_systems) == 1
    assert len(slt.fault_systems[0].branches) == 9
    assert type(slt.fault_systems[0].branches[0]) is Branch  # isinstance() not used to avoid true for inherited classes



#! logic_tree_version_2.py

"""
Define source logic tree structures used in NSHM.
"""
import copy
import json
import pathlib
import warnings
from dataclasses import dataclass, field
from typing import Dict, Iterator, List, Type, Union

import dacite

from nzshm_model.psha_adapter import PshaAdapterInterface

from .. import BranchAttributeValue
from ..core import FaultSystemLogicTreeBase, FaultSystemLogicTreeSpec
from ..version1 import SourceLogicTree as SourceLogicTreeV1


@dataclass
class InversionSource:
    nrml_id: str
    rupture_rate_scaling: Union[float, None] = None  # TODO: needed at this level??
    inversion_id: Union[str, None] = ""
    rupture_set_id: Union[str, None] = ""
    inversion_solution_type: Union[str, None] = ""
    type: str = "inversion"


@dataclass
class DistributedSource:
    nrml_id: str
    rupture_rate_scaling: Union[float, None] = None  # TODO: needed at this level??
    type: str = "distributed"


@dataclass
class Branch:
    values: List[BranchAttributeValue] = field(default_factory=list)
    sources: List[Union[DistributedSource, InversionSource]] = field(default_factory=list)
    weight: float = 1.0
    rupture_rate_scaling: float = 1.0

    def tag(self):
        return str(self.values)


@dataclass
class FaultSystemLogicTree(FaultSystemLogicTreeBase):
    branches: List[Branch] = field(default_factory=list)


@dataclass
class SourceLogicTreeSpec:
    fault_systems: List[FaultSystemLogicTreeSpec] = field(default_factory=list)


@dataclass
class SourceLogicTree:
    version: str
    title: str
    fault_systems: List[FaultSystemLogicTree] = field(default_factory=list)
    logic_tree_version: Union[int, None] = 2

    # correlations: List[SourceLogicTreeCorrelation] = field(
    #     default_factory=list
    # )  # to use for selecting branches and re-weighting when logic trees are correlated

    def derive_spec(self) -> SourceLogicTreeSpec:
        raise NotImplementedError()
        # slt_spec = SourceLogicTreeSpec()
        # for fslt in self.fault_systems:
        #     slt_spec.fault_systems.append(FaultSystemLogicTree.derive_spec(fslt))
        # return slt_spec

    @staticmethod
    def from_dict(data: Dict):
        ltv = data.get("logic_tree_version")
        if not ltv == 2:
            raise ValueError(f"supplied json `logic_tree_version={ltv}` is not supported.")
        return dacite.from_dict(data_class=SourceLogicTree, data=data, config=dacite.Config(strict=True))

    @staticmethod
    def from_json(json_path: Union[pathlib.Path, str]):
        data = json.load(open(json_path))
        return SourceLogicTree.from_dict(data)

    @staticmethod
    def from_source_logic_tree(original_slt: "SourceLogicTreeV1") -> "SourceLogicTree":
        """
        Migrate from old version slt.
        """
        if not isinstance(original_slt, SourceLogicTreeV1):
            raise ValueError(f"supplied object of {type(original_slt)} is not supported.")
        slt = SourceLogicTree(version=original_slt.version, title=original_slt.title)
        for fslt in original_slt.fault_systems:
            new_fslt = FaultSystemLogicTree(short_name=fslt.short_name, long_name=fslt.long_name)
            for branch in fslt.branches:
                # TODO: handle rate scaling
                new_branch = Branch(values=copy.deepcopy(branch.values), weight=branch.weight)
                if branch.onfault_nrml_id:
                    new_branch.sources.append(
                        InversionSource(
                            nrml_id=branch.onfault_nrml_id,
                            inversion_id=branch.inversion_solution_id,
                            rupture_set_id=branch.rupture_set_id,
                        )
                    )
                if branch.distributed_nrml_id:
                    new_branch.sources.append(DistributedSource(nrml_id=branch.distributed_nrml_id))
                new_fslt.branches.append(new_branch)
            slt.fault_systems.append(new_fslt)
        return slt

    @property
    def fault_system_lts(self):
        """
        API alias for fault_systems
        """
        warnings.warn("Please use fault_systems property instead", DeprecationWarning)
        return self.fault_systems

    def psha_adapter(self, provider: Type[PshaAdapterInterface], **kwargs):
        return provider(source_logic_tree=self)

    def __iter__(self):
        self.__current_branch = 0
        self.__branch_list = list(self.__flattened_branches__())
        return self

    def __flattened_branches__(self):
        """
        Produce list of Flattened branches, each with a shallow copy of it's slt and fslt parents
        for use in filtering.

        NB this class is never used for serialising models.
        """
        for fslt in self.fault_systems:
            for branch in fslt.branches:
                fslt_lite = FaultSystemLogicTree(short_name=fslt.short_name, long_name=fslt.long_name)
                slt_lite = SourceLogicTree(
                    title=self.title, version=self.version, logic_tree_version=self.logic_tree_version
                )
                yield FilteredBranch(
                    values=branch.values,
                    sources=copy.deepcopy(branch.sources),
                    weight=branch.weight,
                    fslt=fslt_lite,
                    slt=slt_lite,
                )

    def __next__(self):
        if self.__current_branch >= len(self.__branch_list):
            raise StopIteration
        else:
            self.__current_branch += 1
            return self.__branch_list[self.__current_branch - 1]

    @staticmethod
    def from_branches(branches: Iterator['FilteredBranch']) -> 'SourceLogicTree':
        """
        Build a complete SLT from a iterable od branches.

        We expect that all the branhches have come from a single SLT.
        """

        def match_fslt(slt: SourceLogicTree, fb):
            for fslt in slt.fault_systems:
                if fb.fslt.short_name == fslt.short_name:
                    return fslt

        version = None
        for fb in branches:
            # ensure an slt
            if not version:
                slt = SourceLogicTree(version=fb.slt.version, title=fb.slt.title)
                version = fb.slt.version
            else:
                assert version == fb.slt.version

            # ensure an fslt
            fslt = match_fslt(slt, fb)
            if not fslt:
                fslt = FaultSystemLogicTree(short_name=fb.fslt.short_name, long_name=fb.fslt.long_name)
                slt.fault_systems.append(fslt)
            fslt.branches.append(fb.to_branch())
        return slt


# this should never be serialised, only used for filtering
@dataclass
class FilteredBranch(Branch):
    fslt: 'FaultSystemLogicTree' = FaultSystemLogicTree('shortname', 'longname')
    slt: 'SourceLogicTree' = SourceLogicTree('version', 'title')

    def to_branch(self) -> Branch:
        branch_attributes = {k: v for k, v in self.__dict__.items() if k not in ('fslt', 'slt')}
        return Branch(**branch_attributes)



#! python test_logic_tree.py

from pathlib import Path, PurePath

import pytest

import nzshm_model
from nzshm_model.psha_adapter.openquake import OpenquakeSimplePshaAdapter

MODEL = nzshm_model.get_model_version('NSHM_v1.0.4')
FIXTURE_PATH = Path(__file__).parent / "fixtures"


def test_gmm_logic_tree():

    gmm_logic_tree = MODEL.gmm_logic_tree()

    assert len(gmm_logic_tree.branch_sets) == 3

    # branch set attributes
    assert gmm_logic_tree.branch_sets[0].applyToTectonicRegionType == "Active Shallow Crust"
    assert gmm_logic_tree.branch_sets[0].branchSetID == "bs_crust"
    assert len(gmm_logic_tree.branch_sets[0].branches) == 21

    # branch attributes
    assert gmm_logic_tree.branch_sets[0].branches[0].uncertainty_weight == pytest.approx(0.117000)
    assert gmm_logic_tree.branch_sets[0].branches[0].branchID == "STF22_upper"
    assert gmm_logic_tree.branch_sets[0].branches[0].path() == PurePath('lt1') / "bs_crust" / "STF22_upper"

    # uncertainty model attributes
    assert "Stafford2022" in gmm_logic_tree.branch_sets[0].branches[0].uncertainty_models[0].text
    assert gmm_logic_tree.branch_sets[0].branches[0].uncertainty_models[0].gmpe_name == "[Stafford2022]"
    assert gmm_logic_tree.branch_sets[0].branches[0].uncertainty_models[0].arguments[0] == "mu_branch = \"Upper\""


def test_source_logic_tree():

    slt = MODEL.source_logic_tree()
    src_logic_tree = slt.psha_adapter(provider=OpenquakeSimplePshaAdapter).config()

    assert len(src_logic_tree.branch_sets) == 4

    # branch set attributes
    assert src_logic_tree.branch_sets[0].applyToTectonicRegionType == ""
    assert src_logic_tree.branch_sets[0].branchSetID == "PUY"
    assert src_logic_tree.branch_sets[3].branchSetID == "SLAB"

    assert len(src_logic_tree.branch_sets[0].branches) == 3

    # branch attributes
    assert src_logic_tree.branch_sets[0].branches[0].branchID == "[dm0.7, bN[0.902, 4.6], C4.0, s0.28]"
    assert src_logic_tree.branch_sets[0].branches[0].uncertainty_weight == pytest.approx(0.210000)
    assert (
        src_logic_tree.branch_sets[0].branches[0].path()
        == PurePath('SLT_v9p0p0') / "PUY" / "[dm0.7,bN[0.902,4.6],C4.0,s0.28]"
    )


def test_source_logic_tree_uncertainty_PUY():
    # uncertainty model attributes (SourceUncertaintyModel)
    # NB for crustal, the first is the ltb.onfault_nrml_id
    #    the 2nd is ltb.distributed_nrml_id
    slt = MODEL.source_logic_tree()
    src_logic_tree = slt.psha_adapter(provider=OpenquakeSimplePshaAdapter).config()

    BSID = 0

    assert src_logic_tree.branch_sets[BSID].branchSetID == "PUY"
    assert src_logic_tree.branch_sets[BSID].branches[0].path() == PurePath(
        'SLT_v9p0p0', 'PUY', '[dm0.7,bN[0.902,4.6],C4.0,s0.28]'
    )

    assert len(src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models) == 2

    # ==> NZSHM22_ScaledInversionSolution-QXV0b21hdGlvblRhc2s6MTMyNzM5Mw==_nrml.zip
    # U2NhbGVkSW52ZXJzaW9uU29sdXRpb246MTI5MTM3OA__-ruptures.hdf5
    # U2NhbGVkSW52ZXJzaW9uU29sdXRpb246MTI5MTM3OA__-ruptures_sections.xml
    # U2NhbGVkSW52ZXJzaW9uU29sdXRpb246MTI5MTM3OA__-ruptures.xml
    # http://simple-toshi-ui.s3-website-ap-southeast-2.amazonaws.com/
    #   InversionSolutionNrml/SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE2NTc=
    assert (
        src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models[0].toshi_nrml_id
        == "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE2NTc="
    )

    # ==> npFloor_MULTOT1346GruEEPAScomb-INT_puy_b0.90_N0.73.xml
    # http://simple-toshi-ui.s3-website-ap-southeast-2.amazonaws.com/FileDetail/RmlsZToxMzA3NTM=
    assert src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models[1].toshi_nrml_id == "RmlsZToxMzA3NTM="


def reverse_path(*args):
    return PurePath(*tuple(reversed(PurePath(*args))))


def test_source_logic_tree_uncertainty_SLAB():
    # uncertainty model attributes (SourceUncertaintyModel)
    # NB for slab we have just a distributed model
    slt = MODEL.source_logic_tree()
    src_logic_tree = slt.psha_adapter(provider=OpenquakeSimplePshaAdapter).config()

    BSID = 3

    assert src_logic_tree.branch_sets[BSID].branchSetID == "SLAB"
    assert src_logic_tree.branch_sets[BSID].branches[0].path() == PurePath('SLT_v9p0p0', 'SLAB', '[runiform,d1]')

    assert len(src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models) == 1

    # ==> slab-uniform-1depth-rates.xml.zip
    # slab-uniform-1depth-rates.xml (8.5 MB)
    # http://simple-toshi-ui.s3-website-ap-southeast-2.amazonaws.com/FileDetail/RmlsZToxMjEwMzM=
    assert src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models[0].toshi_nrml_id == "RmlsZToxMjEwMzM="


def test_source_logic_tree_uncertainty_CRU():
    # uncertainty model attributes (SourceUncertaintyModel)
    slt = MODEL.source_logic_tree()
    src_logic_tree = slt.psha_adapter(provider=OpenquakeSimplePshaAdapter).config()

    BSID = 2
    assert src_logic_tree.branch_sets[BSID].branchSetID == "CRU"
    assert src_logic_tree.branch_sets[BSID].branches[0].path() == PurePath(
        'SLT_v9p0p0/CRU/[dmgeodetic,tdFalse,bN[0.823,2.7],C4.2,s0.66]'
    )
    assert len(src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models) == 2

    # ==> NZSHM22_ScaledInversionSolution-QXV0b21hdGlvblRhc2s6MTEzMTQz_nrml.zip
    # http://simple-toshi-ui.s3-website-ap-southeast-2.amazonaws.com/Find/SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE1MDE=
    assert (
        src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models[0].toshi_nrml_id
        == "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE1MDE="
    )

    # ==> npFloor_MULTOT1346GruEEPAScomb-CRU_plyadj_ratestapered_b0.82_N2.57.xml.zip
    # http://simple-toshi-ui.s3-website-ap-southeast-2.amazonaws.com/Find/RmlsZToxMzA3MDc=
    assert src_logic_tree.branch_sets[BSID].branches[0].uncertainty_models[1].toshi_nrml_id == "RmlsZToxMzA3MDc="



#! python test_logic_tree.py

import tempfile
from pathlib import Path

import pytest

from nzshm_model.gmcm_logic_tree import GMCMLogicTree
from nzshm_model.psha_adapter.openquake import OpenquakeSimplePshaAdapter
from nzshm_model.source_logic_tree.version2 import SourceLogicTree

FIXTURE_PATH = Path(__file__).parent.parent.parent / "fixtures"


@pytest.mark.skip("WIP")
def test_source_logic_tree_to_source_xml_basic():

    config = FIXTURE_PATH / 'source_logic_tree_sample_0.json'
    slt = SourceLogicTree.from_json(config)

    print(slt)
    target = FIXTURE_PATH  # noqa
    src_xml = slt.psha_adapter(provider=OpenquakeSimplePshaAdapter).write_config(Path('/tmp/DEMO'))  # noqa
    # assert 0


# @pytest.mark.skip("WIP: need way to compare xml without being exact (or create fixture for exact match)")
def test_gmcm_logic_tree_to_xml():

    # xml_filepath = Path(__file__).parent / 'fixtures' / 'gmcm_logic_tree_example_b.xml'
    json_filepath = Path(__file__).parent / 'fixtures' / 'gmcm_logic_tree_example.json'
    gmcm_logic_tree = GMCMLogicTree.from_json(json_filepath)
    adapter = gmcm_logic_tree.psha_adapter(OpenquakeSimplePshaAdapter)
    # gmcm_logic_tree_expected = adapter.logic_tree_from_xml(xml_filepath)
    # adapter = gmcm_logic_tree
    xml_str = adapter.build_gmcm_xml()
    # gmcm_logic_tree_deserialized = adapter.logic_tree_from_xml(xml_filepath)

    with Path(Path(__file__).parent / 'fixtures' / 'gmcm.xml').open('w') as xmlfile:
        xmlfile.write(adapter.build_gmcm_xml())

    with tempfile.TemporaryDirectory() as tmpdir:
        with open(Path(tmpdir, 'gmcm_lt.xml'), 'w') as xmlfile:
            xmlfile.write(xml_str)
        gmcm_logic_tree_deserialized = adapter.logic_tree_from_xml(Path(tmpdir, 'gmcm_lt.xml'))

    assert gmcm_logic_tree_deserialized == gmcm_logic_tree



from .branch_attribute import BranchAttributeSpec, BranchAttributeValue
from .version1.logic_tree import SourceLogicTree as SourceLogicTreeV1
from .version1.logic_tree import SourceLogicTreeCorrelation
from .version2.logic_tree import SourceLogicTree



"""Script to build an SLT dict/json for K-API from a source_logic_tree model."""

import dataclasses
import json
import logging
import sys

import click

import nzshm_model
from nzshm_model import get_model_version
from nzshm_model.source_logic_tree.logic_tree import SourceLogicTree
from nzshm_model.source_logic_tree.slt_config import from_config, resolve_toshi_source_ids

log = logging.getLogger()
logging.basicConfig(level=logging.WARN)
logging.getLogger('nshm_toshi_client.toshi_client_base').setLevel(logging.INFO)
logging.getLogger('urllib3').setLevel(logging.INFO)
logging.getLogger('botocore').setLevel(logging.INFO)
logging.getLogger('gql.transport.requests').setLevel(logging.WARN)

formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
screen_handler = logging.StreamHandler(stream=sys.stdout)
screen_handler.setFormatter(formatter)
file_handler = logging.FileHandler('slt.log')
file_handler.setLevel(logging.INFO)
file_handler.setFormatter(formatter)
log.addHandler(screen_handler)
log.addHandler(file_handler)


@click.group()
def slt():
    pass


@slt.command(name='ls')
@click.option('-l', '--long', is_flag=True)
def cli_ls(long):
    """List the available model versions."""

    for version in nzshm_model.versions.keys():
        if not long:
            click.echo(version)
            continue
        model = nzshm_model.get_model_version(version)
        click.echo(f"{model.version} `{model.source_logic_tree().title}`")


@slt.command(name='model')
@click.argument('model_id')
@click.option('-B', '--build', is_flag=True)
def cli_model(model_id, build):
    """Get a model by MODEL_ID."""
    model = get_model_version(model_id)
    slt = (
        SourceLogicTree(version="0", title="", fault_system_lts=[model.build_crustal_branches()])
        if build
        else model.source_logic_tree()
    )
    j = json.dumps(dataclasses.asdict(slt), indent=4)
    click.echo(j)


@slt.command(name='from_config')
@click.argument('config_path')
@click.argument('version')
@click.argument('title')
@click.option('-R', '--resolve_toshi_ids', is_flag=True)
def cli_from_config(config_path, version, title, resolve_toshi_ids):
    """Convert a python config file at CONFIG_PATH to an SLT model. Both VERSION and TITLE are required."""

    slt = from_config(config_path, version, title)

    if resolve_toshi_ids:
        slt = resolve_toshi_source_ids(slt)  # get new slt with toshi_ids
    j = json.dumps(dataclasses.asdict(slt), indent=4)
    click.echo(j)


@slt.command(name='spec')
@click.argument('model_id')
@click.option('-B', '--build', is_flag=True)
def cli_model_spec(model_id, build):
    """Get a model specification by MODEL_ID."""
    model = get_model_version(model_id)
    if not model:
        click.echo(f'Oops, the model "{model_id}" was not found', err=True)
        return False
    slt = model.source_logic_tree()
    j = json.dumps(dataclasses.asdict(slt.derive_spec()), indent=4)
    click.echo(j)


if __name__ == '__main__':
    slt()  # pragma: no cover
